@@ -1,12 +1,11 @@
-__version__ = "9.2.101"
+__version__ = "9.2.134"
 
 import logging
-from typing import Set
 
 from .block import Block
-from . import statement as Stmt
-from . import expression as Expr
-from .statement import Assignment
+from . import statement
+from . import expression
+from .statement import Assignment, Statement
 from .expression import Expression, Const, Tmp, Register, UnaryOp, BinaryOp
 from .converter_common import Converter
 from .manager import Manager
@@ -14,8 +13,11 @@ from .block_walker import AILBlockWalker, AILBlockWalkerBase
 
 log = logging.getLogger(__name__)
 
+# REALLY BAD
+Expr = expression
+Stmt = statement
 
-available_converters: Set[str] = set()
+available_converters: set[str] = set()
 
 try:
     from .converter_vex import VEXIRSBConverter
@@ -25,6 +27,7 @@ try:
 except ImportError as e:
     log.debug("Could not import VEXIRSBConverter")
     log.debug(e)
+    VEXIRSBConverter = None
 
 try:
     from .converter_pcode import PCodeIRSBConverter
@@ -34,6 +37,7 @@ try:
 except ImportError as e:
     log.debug("Could not import PCodeIRSBConverter")
     log.debug(e)
+    PCodeIRSBConverter = None
 
 
 class IRSBConverter(Converter):
@@ -58,8 +62,11 @@ class IRSBConverter(Converter):
 __all__ = [
     "available_converters",
     "Block",
+    "expression",
+    "statement",
     "Stmt",
     "Expr",
+    "Statement",
     "Assignment",
     "Expression",
     "Const",
@@ -71,6 +78,6 @@ __all__ = [
     "IRSBConverter",
     "AILBlockWalkerBase",
     "AILBlockWalker",
-    *(["PCodeIRSBConverter"] if "pcode" in available_converters else []),
-    *(["VEXIRSBConverter"] if "vex" in available_converters else []),
+    "PCodeIRSBConverter",
+    "VEXIRSBConverter",
 ]
