@@ -1,8 +1,11 @@
 # pylint:disable=arguments-renamed,isinstance-second-argument-not-valid-type,missing-class-docstring
 from __future__ import annotations
-from enum import IntEnum
+from typing import TYPE_CHECKING, cast
+from collections.abc import Sequence
+from enum import Enum, IntEnum
+from abc import abstractmethod
+from typing_extensions import Self
 
-from typing import TYPE_CHECKING
 
 try:
     import claripy
@@ -21,12 +24,18 @@ class Expression(TaggedObject):
     The base class of all AIL expressions.
     """
 
-    __slots__ = ("depth",)
+    bits: int
+
+    __slots__ = (
+        "bits",
+        "depth",
+    )
 
     def __init__(self, idx, depth, **kwargs):
         super().__init__(idx, **kwargs)
         self.depth = depth
 
+    @abstractmethod
     def __repr__(self):
         raise NotImplementedError()
 
@@ -41,16 +50,18 @@ class Expression(TaggedObject):
             return True
         return type(self) is type(other) and self.likes(other) and self.idx == other.idx
 
-    def likes(self, atom):  # pylint:disable=unused-argument,no-self-use
+    @abstractmethod
+    def likes(self, other):  # pylint:disable=unused-argument,no-self-use
         raise NotImplementedError()
 
-    def matches(self, atom):  # pylint:disable=unused-argument,no-self-use
-        return NotImplementedError()
+    @abstractmethod
+    def matches(self, other):  # pylint:disable=unused-argument,no-self-use
+        raise NotImplementedError()
 
-    def replace(self, old_expr, new_expr):
+    def replace(self, old_expr: Expression, new_expr: Expression) -> tuple[bool, Self]:
         if self is old_expr:
             r = True
-            replaced = new_expr
+            replaced = cast(Self, new_expr)
         elif not isinstance(self, Atom):
             r, replaced = self.replace(old_expr, new_expr)
         else:
@@ -59,10 +70,10 @@ class Expression(TaggedObject):
         return r, replaced
 
     def __add__(self, other):
-        return BinaryOp(None, "Add", [self, other], False, **self.tags)
+        return BinaryOp(None, "Add", [self, other], signed=False, **self.tags)
 
     def __sub__(self, other):
-        return BinaryOp(None, "Sub", [self, other], False, **self.tags)
+        return BinaryOp(None, "Sub", [self, other], signed=False, **self.tags)
 
 
 class Atom(Expression):
@@ -71,25 +82,22 @@ class Atom(Expression):
         "variable_offset",
     )
 
-    def __init__(self, idx, variable=None, variable_offset=0, **kwargs):
+    def __init__(self, idx: int | None, variable=None, variable_offset=0, **kwargs):
         super().__init__(idx, 0, **kwargs)
         self.variable = variable
         self.variable_offset = variable_offset
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "Atom (%d)" % self.idx
 
-    def copy(self):  # pylint:disable=no-self-use
-        return NotImplementedError()
+    def copy(self) -> Self:  # pylint:disable=no-self-use
+        raise NotImplementedError()
 
 
 class Const(Atom):
-    __slots__ = (
-        "value",
-        "bits",
-    )
+    __slots__ = ("value",)
 
-    def __init__(self, idx, variable, value, bits, **kwargs):
+    def __init__(self, idx: int | None, variable, value: int | float, bits: int, **kwargs):
         super().__init__(idx, variable, **kwargs)
 
         self.value = value
@@ -133,12 +141,9 @@ class Const(Atom):
 
 
 class Tmp(Atom):
-    __slots__ = (
-        "tmp_idx",
-        "bits",
-    )
+    __slots__ = ("tmp_idx",)
 
-    def __init__(self, idx, variable, tmp_idx, bits, **kwargs):
+    def __init__(self, idx: int | None, variable, tmp_idx: int, bits, **kwargs):
         super().__init__(idx, variable, **kwargs)
 
         self.tmp_idx = tmp_idx
@@ -168,12 +173,9 @@ class Tmp(Atom):
 
 
 class Register(Atom):
-    __slots__ = (
-        "reg_offset",
-        "bits",
-    )
+    __slots__ = ("reg_offset",)
 
-    def __init__(self, idx, variable, reg_offset, bits, **kwargs):
+    def __init__(self, idx: int | None, variable, reg_offset: int, bits: int, **kwargs):
         super().__init__(idx, variable, **kwargs)
 
         self.reg_offset = reg_offset
@@ -183,10 +185,8 @@ class Register(Atom):
     def size(self):
         return self.bits // 8
 
-    def likes(self, atom):
-        return type(self) is type(atom) and self.reg_offset == atom.reg_offset and self.bits == atom.bits
-
-    matches = likes
+    def likes(self, other):
+        return type(self) is type(other) and self.reg_offset == other.reg_offset and self.bits == other.bits
 
     def __repr__(self):
         return str(self)
@@ -199,6 +199,7 @@ class Register(Atom):
         else:
             return "%s" % str(self.variable.name)
 
+    matches = likes
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
@@ -213,13 +214,13 @@ class VirtualVariableCategory(IntEnum):
     STACK = 1
     MEMORY = 2
     PARAMETER = 3
-    UNKNOWN = 4
+    TMP = 4
+    UNKNOWN = 5
 
 
 class VirtualVariable(Atom):
 
     __slots__ = (
-        "bits",
         "varid",
         "category",
         "oident",
@@ -258,32 +259,40 @@ class VirtualVariable(Atom):
         return self.category == VirtualVariableCategory.PARAMETER
 
     @property
-    def reg_offset(self) -> int | None:
-        if self.was_reg:
-            return self.oident
-        return None
+    def was_tmp(self) -> bool:
+        return self.category == VirtualVariableCategory.TMP
 
     @property
-    def stack_offset(self) -> int | None:
+    def reg_offset(self) -> int:
+        if self.was_reg:
+            return self.oident
+        raise TypeError("Is not a register")
+
+    @property
+    def stack_offset(self) -> int:
         if self.was_stack:
             return self.oident
-        return None
+        raise TypeError("Is not a stack variable")
 
-    def likes(self, atom):
+    @property
+    def tmp_idx(self) -> int | None:
+        return self.oident if self.was_tmp else None
+
+    def likes(self, other):
         return (
-            isinstance(atom, VirtualVariable)
-            and self.varid == atom.varid
-            and self.bits == atom.bits
-            and self.category == atom.category
-            and self.oident == atom.oident
+            isinstance(other, VirtualVariable)
+            and self.varid == other.varid
+            and self.bits == other.bits
+            and self.category == other.category
+            and self.oident == other.oident
         )
 
-    def matches(self, atom):
+    def matches(self, other):
         return (
-            isinstance(atom, VirtualVariable)
-            and self.bits == atom.bits
-            and self.category == atom.category
-            and self.oident == atom.oident
+            isinstance(other, VirtualVariable)
+            and self.bits == other.bits
+            and self.category == other.category
+            and self.oident == other.oident
         )
 
     def __repr__(self):
@@ -315,16 +324,13 @@ class VirtualVariable(Atom):
 
 class Phi(Atom):
 
-    __slots__ = (
-        "bits",
-        "src_and_vvars",
-    )
+    __slots__ = ("src_and_vvars",)
 
     def __init__(
         self,
         idx,
         bits,
-        src_and_vvars: list[tuple[tuple[int, int], VirtualVariable | None]],
+        src_and_vvars: list[tuple[tuple[int, int | None], VirtualVariable | None]],
         **kwargs,
     ):
         super().__init__(idx, **kwargs)
@@ -343,21 +349,21 @@ class Phi(Atom):
     def verbose_op(self) -> str:
         return "Phi"
 
-    def likes(self, atom) -> bool:
-        if isinstance(atom, Phi) and self.bits == atom.bits:
+    def likes(self, other) -> bool:
+        if isinstance(other, Phi) and self.bits == other.bits:
             self_src_and_vvarids = {(src, vvar.varid if vvar is not None else None) for src, vvar in self.src_and_vvars}
             other_src_and_vvarids = {
-                (src, vvar.varid if vvar is not None else None) for src, vvar in atom.src_and_vvars
+                (src, vvar.varid if vvar is not None else None) for src, vvar in other.src_and_vvars
             }
             return self_src_and_vvarids == other_src_and_vvarids
         return False
 
-    def matches(self, atom) -> bool:
-        if isinstance(atom, Phi) and self.bits == atom.bits:
-            if len(self.src_and_vvars) != len(atom.src_and_vvars):
+    def matches(self, other) -> bool:
+        if isinstance(other, Phi) and self.bits == other.bits:
+            if len(self.src_and_vvars) != len(other.src_and_vvars):
                 return False
             self_src_and_vvars = dict(self.src_and_vvars)
-            other_src_and_vvars = dict(atom.src_and_vvars)
+            other_src_and_vvars = dict(other.src_and_vvars)
             for src, self_vvar in self_src_and_vvars.items():
                 if src not in other_src_and_vvars:
                     return False
@@ -381,7 +387,7 @@ class Phi(Atom):
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return stable_hash(("phi", self.bits, tuple(sorted(self.src_and_vvars))))
+        return stable_hash(("phi", self.bits, tuple(sorted(self.src_and_vvars, key=self._src_and_vvar_filter))))
 
     def copy(self) -> Phi:
         return Phi(
@@ -414,6 +420,16 @@ class Phi(Atom):
             )
         return False, self
 
+    @staticmethod
+    def _src_and_vvar_filter(
+        src_and_vvar: tuple[tuple[int, int | None], VirtualVariable | None]
+    ) -> tuple[tuple[int, int], int]:
+        src, vvar = src_and_vvar
+        if src[1] is None:
+            src = src[0], -1
+        vvar_id = vvar.varid if vvar is not None else -1
+        return src, vvar_id
+
 
 class Op(Expression):
     __slots__ = ("op",)
@@ -430,12 +446,20 @@ class Op(Expression):
 class UnaryOp(Op):
     __slots__ = (
         "operand",
-        "bits",
         "variable",
         "variable_offset",
     )
 
-    def __init__(self, idx, op, operand, variable=None, variable_offset=None, bits: int | None = None, **kwargs):
+    def __init__(
+        self,
+        idx: int | None,
+        op: str,
+        operand: Expression,
+        variable=None,
+        variable_offset: int | None = None,
+        bits=None,
+        **kwargs,
+    ):
         super().__init__(idx, (operand.depth if isinstance(operand, Expression) else 0) + 1, op, **kwargs)
 
         self.operand = operand
@@ -457,12 +481,12 @@ class UnaryOp(Op):
             and self.operand.likes(other.operand)
         )
 
-    def matches(self, atom):
+    def matches(self, other):
         return (
-            type(atom) is UnaryOp
-            and self.op == atom.op
-            and self.bits == atom.bits
-            and self.operand.matches(atom.operand)
+            type(other) is UnaryOp
+            and self.op == other.op
+            and self.bits == other.bits
+            and self.operand.matches(other.operand)
         )
 
     __hash__ = TaggedObject.__hash__
@@ -507,10 +531,15 @@ class UnaryOp(Op):
         return self.operand.has_atom(atom, identity=identity)
 
 
-class Convert(UnaryOp):
+class ConvertType(Enum):
     TYPE_INT = 0
     TYPE_FP = 1
 
+
+class Convert(UnaryOp):
+    TYPE_INT = ConvertType.TYPE_INT
+    TYPE_FP = ConvertType.TYPE_FP
+
     __slots__ = (
         "from_bits",
         "to_bits",
@@ -522,13 +551,13 @@ class Convert(UnaryOp):
 
     def __init__(
         self,
-        idx,
-        from_bits,
-        to_bits,
-        is_signed,
-        operand,
-        from_type=TYPE_INT,
-        to_type=TYPE_INT,
+        idx: int | None,
+        from_bits: int,
+        to_bits: int,
+        is_signed: bool,
+        operand: Expression,
+        from_type: ConvertType = TYPE_INT,
+        to_type: ConvertType = TYPE_INT,
         rounding_mode=None,
         **kwargs,
     ):
@@ -719,14 +748,11 @@ class Reinterpret(UnaryOp):
 class BinaryOp(Op):
     __slots__ = (
         "operands",
-        "bits",
-        "signed",
         "variable",
         "variable_offset",
         "floating_point",
         "rounding_mode",
-        "from_bits",  # for divmod
-        "to_bits",  # for divmod
+        "signed",
         "vector_count",
         "vector_size",
     )
@@ -742,7 +768,6 @@ class BinaryOp(Op):
         "MulV": "*",
         "Div": "/",
         "DivF": "/",
-        "DivMod": "/m",
         "Mod": "%",
         "Xor": "^",
         "And": "&",
@@ -759,10 +784,10 @@ class BinaryOp(Op):
         "CmpLE": "<=",
         "CmpGT": ">",
         "CmpGE": ">=",
-        "CmpLTs": "<s",
-        "CmpLEs": "<=s",
-        "CmpGTs": ">s",
-        "CmpGEs": ">=s",
+        "CmpLT (signed)": "<s",
+        "CmpLE (signed)": "<=s",
+        "CmpGT (signed)": ">s",
+        "CmpGE (signed)": ">=s",
         "Concat": "CONCAT",
         "Ror": "ROR",
         "Rol": "ROL",
@@ -778,25 +803,20 @@ class BinaryOp(Op):
         "CmpGE": "CmpLT",
         "CmpLE": "CmpGT",
         "CmpGT": "CmpLE",
-        "CmpLTs": "CmpGEs",
-        "CmpGEs": "CmpLTs",
-        "CmpLEs": "CmpGTs",
-        "CmpGTs": "CmpLEs",
     }
 
     def __init__(
         self,
-        idx,
-        op,
-        operands,
-        signed,
+        idx: int | None,
+        op: str,
+        operands: Sequence[Expression],
+        signed: bool = False,
+        *,
         variable=None,
         variable_offset=None,
         bits=None,
-        floating_point: bool = False,
-        rounding_mode: str | None = None,
-        from_bits: int | None = None,
-        to_bits: int | None = None,
+        floating_point=False,
+        rounding_mode=None,
         vector_count: int | None = None,
         vector_size: int | None = None,
         **kwargs,
@@ -809,11 +829,6 @@ class BinaryOp(Op):
             + 1
         )
 
-        # special handling of initialization with signed op names
-        if op and op.endswith("s"):
-            op = op[:-1]
-            signed = True
-
         super().__init__(idx, depth, op, **kwargs)
 
         assert len(operands) == 2
@@ -849,9 +864,6 @@ class BinaryOp(Op):
         self.vector_count = vector_count
         self.vector_size = vector_size
 
-        self.from_bits = from_bits
-        self.to_bits = to_bits
-
         # TODO: sanity check of operands' sizes for some ops
         # assert self.bits == operands[1].bits
 
@@ -913,14 +925,14 @@ class BinaryOp(Op):
 
         return False
 
-    def replace(self, old_expr, new_expr):
+    def replace(self, old_expr: Expression, new_expr: Expression) -> tuple[bool, BinaryOp]:
         if self.operands[0] == old_expr:
             r0 = True
             replaced_operand_0 = new_expr
         elif isinstance(self.operands[0], Expression):
             r0, replaced_operand_0 = self.operands[0].replace(old_expr, new_expr)
         else:
-            r0, replaced_operand_0 = False, None
+            r0, replaced_operand_0 = False, new_expr
 
         if self.operands[1] == old_expr:
             r1 = True
@@ -928,7 +940,7 @@ class BinaryOp(Op):
         elif isinstance(self.operands[1], Expression):
             r1, replaced_operand_1 = self.operands[1].replace(old_expr, new_expr)
         else:
-            r1, replaced_operand_1 = False, None
+            r1, replaced_operand_1 = False, new_expr
 
         r2, replaced_rm = False, None
         if self.rounding_mode is not None:
@@ -941,12 +953,10 @@ class BinaryOp(Op):
                 self.idx,
                 self.op,
                 [replaced_operand_0 if r0 else self.operands[0], replaced_operand_1 if r1 else self.operands[1]],
-                self.signed,
+                signed=self.signed,
                 bits=self.bits,
                 floating_point=self.floating_point,
                 rounding_mode=replaced_rm if r2 else self.rounding_mode,
-                from_bits=self.from_bits,
-                to_bits=self.to_bits,
                 **self.tags,
             )
         else:
@@ -956,10 +966,10 @@ class BinaryOp(Op):
     def verbose_op(self):
         op = self.op
         if self.floating_point:
-            op += "F"
+            op += " (float)"
         else:
             if self.signed:
-                op += "s"
+                op += " (signed)"
         return op
 
     @property
@@ -971,129 +981,16 @@ class BinaryOp(Op):
             self.idx,
             self.op,
             self.operands[::],
-            self.signed,
             variable=self.variable,
+            signed=self.signed,
             variable_offset=self.variable_offset,
             bits=self.bits,
             floating_point=self.floating_point,
             rounding_mode=self.rounding_mode,
-            from_bits=self.from_bits,
-            to_bits=self.to_bits,
             **self.tags,
         )
 
 
-class TernaryOp(Op):
-    OPSTR_MAP = {}
-
-    __slots__ = (
-        "operands",
-        "bits",
-    )
-
-    def __init__(self, idx, op, operands, bits=None, **kwargs):
-        depth = (
-            max(
-                operands[0].depth if isinstance(operands[0], Expression) else 0,
-                operands[1].depth if isinstance(operands[1], Expression) else 0,
-                operands[2].depth if isinstance(operands[1], Expression) else 0,
-            )
-            + 1
-        )
-        super().__init__(idx, depth, op, **kwargs)
-
-        assert len(operands) == 3
-        self.operands = operands
-        self.bits = bits
-
-    def __str__(self):
-        return f"{self.verbose_op}({self.operands[0]}, {self.operands[1]}, {self.operands[2]})"
-
-    def __repr__(self):
-        return f"{self.verbose_op}({self.operands[0]}, {self.operands[1]}, {self.operands[2]})"
-
-    def likes(self, other):
-        return (
-            type(other) is TernaryOp
-            and self.op == other.op
-            and self.bits == other.bits
-            and is_none_or_likeable(self.operands, other.operands, is_list=True)
-        )
-
-    def matches(self, other):
-        return (
-            type(other) is TernaryOp
-            and self.op == other.op
-            and self.bits == other.bits
-            and is_none_or_matchable(self.operands, other.operands, is_list=True)
-        )
-
-    __hash__ = TaggedObject.__hash__
-
-    def _hash_core(self):
-        return stable_hash((self.op, tuple(self.operands), self.bits))
-
-    def has_atom(self, atom, identity=True):
-        if super().has_atom(atom, identity=identity):
-            return True
-
-        for op in self.operands:
-            if identity and op == atom:
-                return True
-            if not identity and isinstance(op, Atom) and op.likes(atom):
-                return True
-            if isinstance(op, Atom) and op.has_atom(atom, identity=identity):
-                return True
-        return False
-
-    def replace(self, old_expr, new_expr):
-        if self.operands[0] == old_expr:
-            r0 = True
-            replaced_operand_0 = new_expr
-        elif isinstance(self.operands[0], Expression):
-            r0, replaced_operand_0 = self.operands[0].replace(old_expr, new_expr)
-        else:
-            r0, replaced_operand_0 = False, None
-
-        if self.operands[1] == old_expr:
-            r1 = True
-            replaced_operand_1 = new_expr
-        elif isinstance(self.operands[1], Expression):
-            r1, replaced_operand_1 = self.operands[1].replace(old_expr, new_expr)
-        else:
-            r1, replaced_operand_1 = False, None
-
-        if self.operands[2] == old_expr:
-            r2 = True
-            replaced_operand_2 = new_expr
-        elif isinstance(self.operands[2], Expression):
-            r2, replaced_operand_2 = self.operands[2].replace(old_expr, new_expr)
-        else:
-            r2, replaced_operand_2 = False, None
-
-        if r0 or r1 or r2:
-            return True, TernaryOp(
-                self.idx,
-                self.op,
-                [replaced_operand_0, replaced_operand_1, replaced_operand_2],
-                bits=self.bits,
-                **self.tags,
-            )
-        else:
-            return False, self
-
-    @property
-    def verbose_op(self):
-        return self.op
-
-    @property
-    def size(self):
-        return self.bits // 8
-
-    def copy(self) -> TernaryOp:
-        return TernaryOp(self.idx, self.op, self.operands[::], bits=self.bits, **self.tags)
-
-
 class Load(Expression):
     __slots__ = (
         "addr",
@@ -1105,7 +1002,18 @@ class Load(Expression):
         "alt",
     )
 
-    def __init__(self, idx, addr, size, endness, variable=None, variable_offset=None, guard=None, alt=None, **kwargs):
+    def __init__(
+        self,
+        idx: int | None,
+        addr: Expression,
+        size: int,
+        endness: str,
+        variable=None,
+        variable_offset=None,
+        guard=None,
+        alt=None,
+        **kwargs,
+    ):
         depth = max(addr.depth, size.depth if isinstance(size, Expression) else 0) + 1
         super().__init__(idx, depth, **kwargs)
 
@@ -1116,10 +1024,7 @@ class Load(Expression):
         self.alt = alt
         self.variable = variable
         self.variable_offset = variable_offset
-
-    @property
-    def bits(self):
-        return self.size * 8
+        self.bits = self.size * 8
 
     def __repr__(self):
         return str(self)
@@ -1150,6 +1055,7 @@ class Load(Expression):
     def _likes_addr(self, other_addr):
         if hasattr(self.addr, "likes") and hasattr(other_addr, "likes"):
             return self.addr.likes(other_addr)
+
         return self.addr == other_addr
 
     def likes(self, other):
@@ -1201,12 +1107,20 @@ class ITE(Expression):
         "cond",
         "iffalse",
         "iftrue",
-        "bits",
         "variable",
         "variable_offset",
     )
 
-    def __init__(self, idx, cond, iffalse, iftrue, variable=None, variable_offset=None, **kwargs):
+    def __init__(
+        self,
+        idx: int | None,
+        cond: Expression,
+        iffalse: Expression,
+        iftrue: Expression,
+        variable=None,
+        variable_offset=None,
+        **kwargs,
+    ):
         depth = (
             max(
                 cond.depth if isinstance(cond, Expression) else 0,
@@ -1230,22 +1144,22 @@ class ITE(Expression):
     def __str__(self):
         return f"(({self.cond}) ? ({self.iftrue}) : ({self.iffalse}))"
 
-    def likes(self, atom):
+    def likes(self, other):
         return (
-            type(atom) is ITE
-            and self.cond.likes(atom.cond)
-            and self.iffalse.likes(atom.iffalse)
-            and self.iftrue.likes(atom.iftrue)
-            and self.bits == atom.bits
+            type(other) is ITE
+            and self.cond.likes(other.cond)
+            and self.iffalse == other.iffalse
+            and self.iftrue == other.iftrue
+            and self.bits == other.bits
         )
 
-    def matches(self, atom):
+    def matches(self, other):
         return (
-            type(atom) is ITE
-            and self.cond.matches(atom.cond)
-            and self.iffalse.matches(atom.iffalse)
-            and self.iftrue.matches(atom.iftrue)
-            and self.bits == atom.bits
+            type(other) is ITE
+            and self.cond.matches(other.cond)
+            and self.iffalse == other.iffalse
+            and self.iftrue == other.iftrue
+            and self.bits == other.bits
         )
 
     __hash__ = TaggedObject.__hash__
@@ -1299,69 +1213,168 @@ class ITE(Expression):
 
 class DirtyExpression(Expression):
     __slots__ = (
-        "dirty_expr",
-        "bits",
+        "callee",
+        "guard",
+        "operands",
+        "mfx",
+        "maddr",
+        "msize",
     )
 
-    def __init__(self, idx, dirty_expr, bits=None, **kwargs):
+    def __init__(
+        self,
+        idx,
+        callee: str,
+        operands: list[Expression],
+        *,
+        guard: Expression | None = None,
+        mfx: str | None = None,
+        maddr: Expression | None = None,
+        msize: int | None = None,
+        # TODO: fxstate (guest state effects) is not modeled yet
+        bits=None,
+        **kwargs,
+    ):
         super().__init__(idx, 1, **kwargs)
-        self.dirty_expr = dirty_expr
+
+        self.callee = callee
+        self.guard = guard
+        self.operands = operands
+        self.mfx = mfx
+        self.maddr = maddr
+        self.msize = msize
         self.bits = bits
 
+    @property
+    def op(self) -> str:
+        return self.callee
+
+    @property
+    def verbose_op(self) -> str:
+        return self.op
+
     def likes(self, other):
-        return type(other) is DirtyExpression and other.dirty_expr == self.dirty_expr
+        return (
+            type(other) is DirtyExpression
+            and other.callee == self.callee
+            and is_none_or_likeable(other.guard, self.guard)
+            and len(self.operands) == len(other.operands)
+            and all(op1.likes(op2) for op1, op2 in zip(self.operands, other.operands))
+            and other.mfx == self.mfx
+            and is_none_or_likeable(other.maddr, self.maddr)
+            and other.msize == self.msize
+            and self.bits == other.bits
+        )
+
+    def matches(self, other):
+        return (
+            type(other) is DirtyExpression
+            and other.callee == self.callee
+            and is_none_or_matchable(other.guard, self.guard)
+            and len(self.operands) == len(other.operands)
+            and all(op1.matches(op2) for op1, op2 in zip(self.operands, other.operands))
+            and other.mfx == self.mfx
+            and is_none_or_matchable(other.maddr, self.maddr)
+            and other.msize == self.msize
+            and self.bits == other.bits
+        )
 
-    matches = likes
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return stable_hash((DirtyExpression, self.dirty_expr))
+        return stable_hash(
+            (
+                DirtyExpression,
+                self.callee,
+                self.guard,
+                tuple(self.operands),
+                self.mfx,
+                self.maddr,
+                self.msize,
+                self.bits,
+            )
+        )
 
     def __repr__(self):
-        return "DirtyExpression (%s)" % type(self.dirty_expr)
+        return f"[D] {self.callee}({', '.join(repr(op) for op in self.operands)})"
 
     def __str__(self):
-        return "[D] %s" % str(self.dirty_expr)
+        return f"[D] {self.callee}({', '.join(repr(op) for op in self.operands)})"
 
     def copy(self) -> DirtyExpression:
-        return DirtyExpression(self.idx, self.dirty_expr, bits=self.bits, **self.tags)
+        return DirtyExpression(
+            self.idx,
+            self.callee,
+            self.operands,
+            guard=self.guard,
+            mfx=self.mfx,
+            maddr=self.maddr,
+            msize=self.msize,
+            bits=self.bits,
+            **self.tags,
+        )
 
-    def replace(self, old_expr, new_expr):
-        if old_expr is self.dirty_expr:
-            return True, DirtyExpression(self.idx, new_expr, bits=self.bits, **self.tags)
+    def replace(self, old_expr: Expression, new_expr: Expression):
+        new_operands = []
+        replaced = False
+        for op in self.operands:
+            if old_expr == op:
+                replaced = True
+                new_operands.append(new_expr)
+            else:
+                r, new_op = op.replace(old_expr, new_expr)
+                if r:
+                    replaced = True
+                    new_operands.append(new_op)
+                else:
+                    new_operands.append(op)
 
-        if isinstance(self.dirty_expr, Expression):
-            replaced, new_dirty_expr = self.dirty_expr.replace(old_expr, new_expr)
-        else:
-            replaced = False
-            new_dirty_expr = None
         if replaced:
-            return True, DirtyExpression(self.idx, new_dirty_expr, bits=self.bits, **self.tags)
+            return True, DirtyExpression(
+                self.idx,
+                self.callee,
+                new_operands,
+                guard=self.guard,
+                mfx=self.mfx,
+                maddr=self.maddr,
+                msize=self.msize,
+                bits=self.bits,
+                **self.tags,
+            )
         else:
             return False, self
 
     @property
     def size(self):
+        if self.bits is None:
+            return None
         return self.bits // 8
 
 
 class VEXCCallExpression(Expression):
     __slots__ = (
-        "cee_name",
+        "callee",
         "operands",
-        "bits",
     )
 
-    def __init__(self, idx, cee_name, operands, bits=None, **kwargs):
+    def __init__(self, idx: int | None, callee: str, operands: tuple[Expression, ...], bits: int, **kwargs):
         super().__init__(idx, max(operand.depth for operand in operands), **kwargs)
-        self.cee_name = cee_name
+        self.callee = callee
         self.operands = operands
         self.bits = bits
 
+    @property
+    def op(self) -> str:
+        return self.callee
+
+    @property
+    def verbose_op(self) -> str:
+        return self.op
+
     def likes(self, other):
         return (
             type(other) is VEXCCallExpression
-            and other.cee_name == self.cee_name
+            and other.callee == self.callee
             and len(self.operands) == len(other.operands)
             and self.bits == other.bits
             and all(op1.likes(op2) for op1, op2 in zip(other.operands, self.operands))
@@ -1370,7 +1383,7 @@ class VEXCCallExpression(Expression):
     def matches(self, other):
         return (
             type(other) is VEXCCallExpression
-            and other.cee_name == self.cee_name
+            and other.callee == self.callee
             and len(self.operands) == len(other.operands)
             and self.bits == other.bits
             and all(op1.matches(op2) for op1, op2 in zip(other.operands, self.operands))
@@ -1379,17 +1392,17 @@ class VEXCCallExpression(Expression):
     __hash__ = TaggedObject.__hash__
 
     def _hash_core(self):
-        return stable_hash((VEXCCallExpression, self.cee_name, self.bits, tuple(self.operands)))
+        return stable_hash((VEXCCallExpression, self.callee, self.bits, tuple(self.operands)))
 
     def __repr__(self):
-        return f"VEXCCallExpression [{self.cee_name}]"
+        return f"VEXCCallExpression [{self.callee}({', '.join(repr(op) for op in self.operands)})]"
 
     def __str__(self):
         operands_str = ", ".join(repr(op) for op in self.operands)
-        return f"{self.cee_name}({operands_str})"
+        return f"{self.callee}({operands_str})"
 
     def copy(self) -> VEXCCallExpression:
-        return VEXCCallExpression(self.idx, self.cee_name, self.operands, bits=self.bits, **self.tags)
+        return VEXCCallExpression(self.idx, self.callee, self.operands, bits=self.bits, **self.tags)
 
     def replace(self, old_expr, new_expr):
         new_operands = []
@@ -1407,12 +1420,14 @@ class VEXCCallExpression(Expression):
                     new_operands.append(operand)
 
         if replaced:
-            return True, VEXCCallExpression(self.idx, self.cee_name, tuple(new_operands), bits=self.bits, **self.tags)
+            return True, VEXCCallExpression(self.idx, self.callee, list(new_operands), bits=self.bits, **self.tags)
         else:
             return False, self
 
     @property
     def size(self):
+        if self.bits is None:
+            return None
         return self.bits // 8
 
 
@@ -1430,6 +1445,7 @@ class MultiStatementExpression(Expression):
         super().__init__(idx, expr.depth + 1, **kwargs)
         self.stmts = stmts
         self.expr = expr
+        self.bits = self.expr.bits
 
     __hash__ = TaggedObject.__hash__
 
@@ -1444,12 +1460,12 @@ class MultiStatementExpression(Expression):
             and self.expr.likes(other.expr)
         )
 
-    def matches(self, atom):
+    def matches(self, other):
         return (
-            type(self) is type(atom)
-            and len(self.stmts) == len(atom.stmts)
-            and all(s_stmt.matches(o_stmt) for s_stmt, o_stmt in zip(self.stmts, atom.stmts))
-            and self.expr.matches(atom.expr)
+            type(self) is type(other)
+            and len(self.stmts) == len(other.stmts)
+            and all(s_stmt.matches(o_stmt) for s_stmt, o_stmt in zip(self.stmts, other.stmts))
+            and self.expr.matches(other.expr)
         )
 
     def __repr__(self):
@@ -1461,10 +1477,6 @@ class MultiStatementExpression(Expression):
         concatenated_str = ", ".join(stmts_str + [expr_str])
         return f"({concatenated_str})"
 
-    @property
-    def bits(self):
-        return self.expr.bits
-
     @property
     def size(self):
         return self.expr.size
@@ -1502,14 +1514,22 @@ class MultiStatementExpression(Expression):
 
 class BasePointerOffset(Expression):
     __slots__ = (
-        "bits",
         "base",
         "offset",
         "variable",
         "variable_offset",
     )
 
-    def __init__(self, idx, bits, base, offset, variable=None, variable_offset=None, **kwargs):
+    def __init__(
+        self,
+        idx: int | None,
+        bits: int,
+        base: Expression | str,
+        offset: int,
+        variable=None,
+        variable_offset=None,
+        **kwargs,
+    ):
         super().__init__(idx, (offset.depth if isinstance(offset, Expression) else 0) + 1, **kwargs)
         self.bits = bits
         self.base = base
@@ -1570,7 +1590,7 @@ class BasePointerOffset(Expression):
 class StackBaseOffset(BasePointerOffset):
     __slots__ = ()
 
-    def __init__(self, idx, bits, offset, **kwargs):
+    def __init__(self, idx: int | None, bits: int, offset: int, **kwargs):
         # stack base offset is always signed
         if offset >= (1 << (bits - 1)):
             offset -= 1 << bits
@@ -1589,7 +1609,7 @@ def negate(expr: Expression) -> Expression:
             expr.idx,
             BinaryOp.COMPARISON_NEGATION[expr.op],
             expr.operands,
-            expr.signed,
+            signed=expr.signed,
             bits=expr.bits,
             floating_point=expr.floating_point,
             rounding_mode=expr.rounding_mode,
